/*
 * TclMilter - Tcl Interface to Sendmail's Milter (Mail Filter) API
 * Copyright (C) 2005 Muonics & Michael Kirkham <mikek@muonics.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <tcl.h>
#include <netinet/in.h>
#include "libmilter/mfapi.h"

/*
 * Global mutex to prevent thread collisions.  Maybe not strictly
 * necessary, since only one thread should be able to do anything
 * beyond stopping the milter.
 */

static Tcl_Mutex	gsMutex				= NULL;

/*
 * The primary interpreter - the one in which TclMilter was first loaded.
 * This is the interpreter that callbacks are invoked in, and all data
 * and such is tied to this interpreter.  Other interpreters may only
 * stop the milter.
 */

static Tcl_Interp*	gpInterp			= NULL;

/*
 * Registered Tcl callbacks to invoke by Milter callbacks.
 */

static Tcl_Obj*     gspObjCmdConnect	= NULL;
static Tcl_Obj*     gspObjCmdHelo		= NULL;
static Tcl_Obj*     gspObjCmdEnvfrom	= NULL;
static Tcl_Obj*     gspObjCmdEnvrcpt	= NULL;
static Tcl_Obj*     gspObjCmdHeader		= NULL;
static Tcl_Obj*     gspObjCmdEoh		= NULL;
static Tcl_Obj*     gspObjCmdBody		= NULL;
static Tcl_Obj*     gspObjCmdEom		= NULL;
static Tcl_Obj*     gspObjCmdAbort		= NULL;
static Tcl_Obj*     gspObjCmdClose		= NULL;

/*
 * Milter Configuration
 */

static Tcl_Obj*		gspObjMilterName	= NULL;
static Tcl_Obj*		gspObjConnection	= NULL;
static int			gsTimeout			= 7210;
static int			gsBacklog			= -1;
static int			gsDebugLevel		= 0;

static int			TM_CtxObjCmd(ClientData clientData, Tcl_Interp* interp,
						int objc, Tcl_Obj* CONST objv[]);

#define SetObjResult(interp, pObj) \
	Tcl_SetObjResult(interp, (pObj) ? pObj : Tcl_NewObj());

#ifndef bool
# define bool	int
# define TRUE	1
# define FALSE	0
#endif /* ! bool */

/*
 * Storage structure used as the opaque private data pointer associated
 * with mitler message contexts.  Each context is given its own Tcl
 * hash table for arbitrary associated-data storage by Tcl scripts.
 */

typedef struct TM_PrivData
{
	Tcl_Command		mToken;		/* Tcl command to access context */
	Tcl_HashTable	mData;		/* Custom private data */
	SMFICTX*		mpContext;	/* Reference back to the context its self */
} TM_PrivData;


/*
 *-------------------------------------------------------------------------
 *
 * TM_EvalObj --
 *
 *	This procedure is a wrapper around Tcl_EvalObj, which allows the
 *	Tcl_Obj pointer to be NULL and converts Tcl return values to
 *	libmilter return codes.  It is invoked by each of the callbacks
 *	to invoke the Tcl handler for the callback.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 * 	Depends on the Tcl script.  An invalid return value will result in
 * 	the message being rejected with a temporary failure.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_EvalObj(SMFICTX* pCtx, Tcl_Interp* interp, Tcl_Obj* pObjCmd)
{
	static const char* ppCodes[] =
	{
		"SMFIS_CONTINUE", "SMFIS_REJECT", "SMFIS_DISCARD",
		"SMFIS_ACCEPT", "SMFIS_TEMPFAIL",
		(char *)NULL
	};

	enum retCodes
	{
		RET_SMFIS_CONTINUE, RET_SMFIS_REJECT, RET_SMFIS_DISCARD,
		RET_SMFIS_ACCEPT, RET_SMFIS_TEMPFAIL,
	};

	int retCode;

	/*
	 * Treat a NULL object pointer as a no-op.  Otherwise, evaluate
	 * the given object (generally a handler for one of the milter
	 * callbacks).
	 */

	if (!pObjCmd)
		return SMFIS_CONTINUE;

	if (Tcl_EvalObj(interp, pObjCmd) == TCL_ERROR)
		return SMFIS_TEMPFAIL;

	/*
	 * Convert the return value provided by the script into an expected
	 * return code for the milter API.
	 */

	if (Tcl_GetIndexFromObj(interp, Tcl_GetObjResult(interp),
			ppCodes, "sfsistat", 0, &retCode) == TCL_ERROR)
	{
		smfi_setreply(pCtx, "451", "4.3.0", "TclMitler: handler return code");
		return SMFIS_TEMPFAIL;
	}

	switch ((enum retCodes)retCode)
	{
	case RET_SMFIS_CONTINUE:
		retCode = SMFIS_CONTINUE;
		break;
	case RET_SMFIS_REJECT:
		retCode = SMFIS_REJECT;
		break;
	case RET_SMFIS_DISCARD:
		retCode = SMFIS_DISCARD;
		break;
	case RET_SMFIS_ACCEPT:
		retCode = SMFIS_ACCEPT;
		break;
	case RET_SMFIS_TEMPFAIL:
		retCode = SMFIS_TEMPFAIL;
		break;

	default:
		/* This shouldn't ever happen. */
		smfi_setreply(pCtx, "451", "4.3.0", "TclMilter: internal error");
		retCode = SMFIS_CONTINUE;
	};

	return retCode;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_cleanup --
 *
 *	This procedure is invoked by libmilter after message processing is
 *	complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The Tcl command associated with the context is deleted, along with
 *	the private data storage in the hash table.
 *
 *-------------------------------------------------------------------------
 */

static void
TM_cleanup(SMFICTX* pCtx, bool deleteCtxCmd)
{
	TM_PrivData*	pPriv	= NULL;

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	/* Free up resources stashed in the hash table. */

	if (pPriv)
	{
		Tcl_HashSearch	status;
		Tcl_HashEntry*	pEnt = NULL;

		/* Decrement refcounts of each Tcl_Obj stored in the table */

		pEnt = Tcl_FirstHashEntry(&pPriv->mData, &status);

		while (pEnt)
		{
			Tcl_DecrRefCount((Tcl_Obj*)Tcl_GetHashValue(pEnt));
			pEnt = Tcl_NextHashEntry(&status);
		}

		/* Delete the hash table its self. */

		Tcl_DeleteHashTable(&pPriv->mData);

		/* Delete the command associated with the context */

		if (deleteCtxCmd)
			Tcl_DeleteCommandFromToken(gpInterp, pPriv->mToken);

		smfi_setpriv(pCtx, NULL);

		/* Finally, free the private data structure */

		ckfree((char *)pPriv);
	}

	Tcl_MutexUnlock(&gsMutex);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_CtxObjCmdDeleted --
 *
 *	This procedure is invoked by Tcl when a script deletes the command
 *	associated with the context.  Afterwards, it's no longer appropriate
 *	to invoke Tcl handlers for the context.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	See TM_cleanup
 *
 *-------------------------------------------------------------------------
 */
static void
TM_CtxObjCmdDeleted(ClientData clientData)
{
	TM_cleanup((SMFICTX *)clientData, FALSE);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_connect --
 *
 *	This procedure is invoked by libmilter whenever a new connection is
 *	established due to incoming mail.  The registered Tcl handler, if
 *	any, is invoked with the host name and address for the connected client.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_connect(SMFICTX* pCtx, char* pHostName, _SOCK_ADDR* pHostAddr)
{
	TM_PrivData*	pPriv		= NULL;
	Tcl_Obj*		pObjCmd		= NULL;
	char			pCtxStr[32];

	/* Initialize private data structure for the new context */

	pPriv = (TM_PrivData*)ckalloc(sizeof(TM_PrivData));

	if (!pPriv)
		return SMFIS_TEMPFAIL;

	Tcl_MutexLock(&gsMutex);

	Tcl_InitObjHashTable(&pPriv->mData);

	/* Set private data pointer to the prepared hash table. */

	smfi_setpriv(pCtx, pPriv);

	/* Create Tcl command for interacting with the context. */

	snprintf(pCtxStr, sizeof(pCtxStr)-1, "%p", pCtx);
	pPriv -> mToken = Tcl_CreateObjCommand(gpInterp, pCtxStr,
		TM_CtxObjCmd, (ClientData)pCtx, TM_CtxObjCmdDeleted);

	/* Invoke the registered connect callback, if any */

	if (gspObjCmdConnect)
	{
		Tcl_Obj*	pObjAddr = Tcl_NewObj();

		pObjCmd = Tcl_DuplicateObj(gspObjCmdConnect);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pCtxStr, -1));
		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pHostName, -1));

		if (pHostAddr && (AF_INET == pHostAddr -> sa_family))
		{
			struct sockaddr_in* sin = (struct sockaddr_in*) pHostAddr;

			Tcl_ListObjAppendElement(gpInterp, pObjAddr,
				Tcl_NewStringObj("address", 7));
			Tcl_ListObjAppendElement(gpInterp, pObjAddr,
				Tcl_NewWideIntObj(sin->sin_addr.s_addr));
			Tcl_ListObjAppendElement(gpInterp, pObjAddr,
				Tcl_NewStringObj("port", 4));
			Tcl_ListObjAppendElement(gpInterp, pObjAddr,
				Tcl_NewIntObj(sin->sin_port));
		}

		Tcl_ListObjAppendElement(gpInterp, pObjCmd, pObjAddr);
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_helo --
 *
 *	This procedure is invoked by libmilter in response to the HELO/EHLO
 *	SMTP command. The registered Tcl handler, if any, is invoked with the
 *	provided string passed with the HELO command.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_helo(SMFICTX* pCtx, char* pHeloHost)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdHelo)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdHelo);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pHeloHost, -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_envfrom --
 *
 *	This procedure is invoked by libmilter in response to the MAIL FROM
 *	SMTP command. The registered Tcl handler, if any, is invoked with
 *	the list of addresses provided to the command.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_envfrom(SMFICTX* pCtx, char** argv)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdEnvfrom)
	{
		Tcl_Obj*	pObjList	= Tcl_NewObj();

		pObjCmd = Tcl_DuplicateObj(gspObjCmdEnvfrom);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));

		while (argv && *argv)
		{
			Tcl_ListObjAppendElement(gpInterp, pObjList,
				Tcl_NewStringObj(*argv, -1));

			argv++;
		}

		Tcl_ListObjAppendElement(gpInterp, pObjCmd, pObjList);
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_envrcpt --
 *
 *	This procedure is invoked by libmilter in response to the RCPT TO
 *	SMTP command. The registered Tcl handler, if any, is invoked with
 *	the list of addresses provided to the command.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_envrcpt(SMFICTX* pCtx, char** argv)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdEnvrcpt)
	{
		Tcl_Obj*	pObjList	= Tcl_NewObj();

		pObjCmd = Tcl_DuplicateObj(gspObjCmdEnvrcpt);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));

		while (argv && *argv)
		{
			Tcl_ListObjAppendElement(gpInterp, pObjList,
				Tcl_NewStringObj(*argv, -1));

			argv++;
		}

		Tcl_ListObjAppendElement(gpInterp, pObjCmd, pObjList);
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_header --
 *
 *	This procedure is invoked by libmilter for each header supplied by
 *	an incoming message.  The registered Tcl handler, if any, is invoked with
 *	the name of the header and its contents.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_header(SMFICTX* pCtx, char* pHdrName, char* pHdrValue)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdHeader)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdHeader);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pHdrName, -1));
		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pHdrValue, -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_eoh --
 *
 *	This procedure is invoked by libmilter once all headers have been
 *	sent and passed through TM_header, to signify the end of headers.
 *	The registered Tcl handler, if any, is invoked.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_eoh(SMFICTX* pCtx)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdEoh)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdEoh);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_body --
 *
 *	This procedure is invoked by libmilter once for each "chunk" of body
 *	text in the message.  The registered Tcl handler, if any, is invoked
 *	with the contents of the body chunk.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_body(SMFICTX* pCtx, unsigned char* pBody, size_t bodySz)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdBody)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdBody);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(pBody, bodySz));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_eom --
 *
 *	This procedure is invoked by libmilter at the end of the message, after
 *	all body chunks have been handled.  The registered Tcl handler, if any,
 *	is invoked.  Most or all message rewriting activities - such as
 *	adding/removing headers or recipients - must be performed within this
 *	callback.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_eom(SMFICTX* pCtx)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdEom)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdEom);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_abort --
 *
 *	This procedure is invoked by libmilter if processing of the message
 *	is to be aborted.  Libmilter will call either this function -or-
 *	TM_close, but not both. The registered Tcl handler, if any, is invoked.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_abort(SMFICTX* pCtx)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdAbort)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdAbort);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	return TM_EvalObj(pCtx, gpInterp, pObjCmd);
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_close --
 *
 *	This procedure is invoked by libmilter if processing is finished and
 *	the connection has been closed.  Libmilter will call either this
 *	function -or- TM_abort, but not both. The registered Tcl handler,
 *	if any, is invoked.
 *
 * Results:
 * 	SMFIS_CONTINUE
 * 	SMFIS_REJECT
 * 	SMFIS_DISCARD
 * 	SMFIS_ACCEPT
 * 	SMFIS_TEMPFAIL
 *
 * Side effects:
 *	Depends on the Tcl script.
 *
 *-------------------------------------------------------------------------
 */

static sfsistat
TM_close(SMFICTX* pCtx)
{
	TM_PrivData*	pPriv	= NULL;
	Tcl_Obj*		pObjCmd	= NULL;
	int				result;

	/* Invoke the registered connect callback, if any */

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv && gspObjCmdClose)
	{
		pObjCmd = Tcl_DuplicateObj(gspObjCmdClose);

		Tcl_ListObjAppendElement(gpInterp, pObjCmd,
			Tcl_NewStringObj(Tcl_GetCommandName(gpInterp, pPriv->mToken), -1));
	}

	Tcl_MutexUnlock(&gsMutex);

	result = TM_EvalObj(pCtx, gpInterp, pObjCmd);
	TM_cleanup(pCtx, TRUE);
	return result;
}


/*
 * The milter description structure, which is used to set up callbacks
 * and behavior for the milter.  Each of the callbacks that libmilter
 * its self calls are wrappers around calls to Tcl scripts.  The only
 * thing configurable here is the filter name, and only prior to
 * calling "milter main".
 */

struct smfiDesc smfilter =
{
	"TclMilter",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */

					/* flags - controls what actions are supported */

	(SMFIF_ADDHDRS | SMFIF_CHGHDRS | SMFIF_CHGBODY |
		SMFIF_ADDRCPT | SMFIF_DELRCPT | SMFIF_QUARANTINE),

	TM_connect,		/* connection info handler */
	TM_helo,		/* SMTP HELO command handler */
	TM_envfrom,		/* envelope sender handler */
	TM_envrcpt,		/* envelope recipient handler */
	TM_header,		/* header handler */
	TM_eoh,			/* end of header handler */
	TM_body,		/* body block handler */
	TM_eom,			/* end of message handler */
	TM_abort,		/* message aborted handler */
	TM_close		/* connection cleanup handler */
};


/*
 *-------------------------------------------------------------------------
 *
 * TM_ConfigObjCmd --
 *
 *	This procedure is a handler for the "milter configure" Tcl command.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	Milter configuration - such as timeout and log level - may be changed.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_ConfigObjCmd
(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	static const char* ppCmdOptions[] =
	{
		"-name", "-connection", "-timeout", "-backlog", "-debuglevel",
		(char *)NULL
	};

	enum cmdOptions
	{
		OPT_NAME, OPT_CONNECTION, OPT_TIMEOUT, OPT_BACKLOG, OPT_DEBUG
	};

	int optIndex;

	objc--; objv++;

	if (!objc)
	{
		Tcl_Obj* pObjResult = Tcl_NewObj();

		Tcl_MutexLock(&gsMutex);

		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewStringObj("-name", 5));
		Tcl_ListObjAppendElement(interp, pObjResult,
			gspObjMilterName);
		
		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewStringObj("-connection", 11));
		Tcl_ListObjAppendElement(interp, pObjResult,
			gspObjConnection);

		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewStringObj("-timeout", 8));
		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewIntObj(gsTimeout));

		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewStringObj("-backlog", 8));
		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewIntObj(gsBacklog));

		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewStringObj("-debuglevel", 11));
		Tcl_ListObjAppendElement(interp, pObjResult,
			Tcl_NewIntObj(gsDebugLevel));

		Tcl_MutexUnlock(&gsMutex);

		Tcl_SetObjResult(interp, pObjResult);

		return TCL_OK;
	}

	if (Tcl_GetIndexFromObj(interp, objv[0], ppCmdOptions, "event",
			0, &optIndex) == TCL_ERROR)
		return TCL_ERROR;

	objc--; objv++;

	if (!objc)
	{
		Tcl_MutexLock(&gsMutex);

		switch ((enum cmdOptions) optIndex)
		{
		case OPT_NAME:
			Tcl_SetObjResult(interp, gspObjMilterName);
			break;
		case OPT_CONNECTION:
			Tcl_SetObjResult(interp, gspObjConnection);
			break;
		case OPT_TIMEOUT:
			Tcl_SetObjResult(interp, Tcl_NewIntObj(gsTimeout));
			break;
		case OPT_BACKLOG:
			Tcl_SetObjResult(interp, Tcl_NewIntObj(gsBacklog));
			break;
		case OPT_DEBUG:
			Tcl_SetObjResult(interp, Tcl_NewIntObj(gsDebugLevel));
			break;
		default:
			Tcl_MutexUnlock(&gsMutex);
			Tcl_SetResult(interp, "Internal error - option not supported",
				TCL_STATIC);
			return TCL_ERROR;
		}

		Tcl_MutexUnlock(&gsMutex);
		return TCL_OK;
	}

	switch ((enum cmdOptions) optIndex)
	{
	case OPT_NAME:
		if (Tcl_GetCharLength(objv[0]) == 0)
		{
			Tcl_SetResult(interp, "-name value must not be zero-length",
				TCL_STATIC);
			return TCL_ERROR;
		}

		Tcl_MutexLock(&gsMutex);
		Tcl_IncrRefCount(objv[0]);
		Tcl_DecrRefCount(gspObjMilterName);
		gspObjMilterName = objv[0];
		Tcl_MutexUnlock(&gsMutex);
		break;

	case OPT_CONNECTION:
		Tcl_MutexLock(&gsMutex);

		Tcl_IncrRefCount(objv[0]);
		Tcl_DecrRefCount(gspObjConnection);
		gspObjConnection = objv[0];

		smfi_setconn(Tcl_GetString(gspObjConnection));

		Tcl_MutexUnlock(&gsMutex);
		break;

	case OPT_TIMEOUT:
		{
			int timeout;

			if (Tcl_GetIntFromObj(interp, objv[0], &timeout) == TCL_ERROR)
				return TCL_ERROR;

			if (timeout <= 0)
			{
				Tcl_SetResult(interp, "-timeout must be greater than 0",
					TCL_STATIC);
				return TCL_ERROR;
			}

			Tcl_MutexLock(&gsMutex);

			gsTimeout = timeout;
			smfi_settimeout(timeout);

			Tcl_MutexUnlock(&gsMutex);
		}
		break;

	case OPT_BACKLOG:
		{
			int backlog;

			if (Tcl_GetIntFromObj(interp, objv[0], &backlog) == TCL_ERROR)
				return TCL_ERROR;

			if (backlog <= 0)
			{
				Tcl_SetResult(interp, "-backlog must be greater than 0",
					TCL_STATIC);
				return TCL_ERROR;
			}

			Tcl_MutexLock(&gsMutex);

			gsBacklog = backlog;
			smfi_setbacklog(backlog);

			Tcl_MutexUnlock(&gsMutex);
		}
		break;

	case OPT_DEBUG:
		{
			int level;

			if (Tcl_GetIntFromObj(interp, objv[0], &level) == TCL_ERROR)
				return TCL_ERROR;

			if (level < 0)
			{
				Tcl_SetResult(interp,
					"-debuglevel must be greater than or equal to 0",
					TCL_STATIC);
				return TCL_ERROR;
			}

			Tcl_MutexLock(&gsMutex);

			gsDebugLevel = level;
			smfi_setdbg(level);

			Tcl_MutexUnlock(&gsMutex);
		}
		break;

	default:
		Tcl_SetResult(interp, "Internal error - option not supported",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_RegisterObjCmd --
 *
 *	This procedure is a handler for the "milter register" Tcl command to
 *	specify (or query) the procedures that are to be invoked by each of
 *	the libmilter callbacks at different stages of message processing.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	None.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_RegisterObjCmd
(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	static const char* ppCmdOptions[] =
	{
		"connect", "helo", "envfrom", "envrcpt", "header",
		"eoh", "body", "eom", "abort", "close",
		(char *)NULL
	};

	enum cmdOptions
	{
		OPT_CONNECT, OPT_HELO, OPT_ENVFROM, OPT_ENVRCPT, OPT_HEADER,
		OPT_EOH, OPT_BODY, OPT_EOM, OPT_ABORT, OPT_CLOSE
	};

	int optIndex;

	if ((objc < 2) || (objc > 3))
	{
		Tcl_WrongNumArgs(interp, 1, objv, "event ?script?");
		return TCL_ERROR;
	}

	objc--; objv++;

	if (Tcl_GetIndexFromObj(interp, objv[0], ppCmdOptions, "event",
			0, &optIndex) == TCL_ERROR)
		return TCL_ERROR;

	objc--; objv++;

	/* Return the current script if none is registered */

	if (!objc)
	{
		Tcl_MutexLock(&gsMutex);

		switch ((enum cmdOptions) optIndex)
		{
		case OPT_CONNECT:
			SetObjResult(interp, gspObjCmdConnect);
			break;
		case OPT_HELO:
			SetObjResult(interp, gspObjCmdHelo);
			break;
		case OPT_ENVFROM:
			SetObjResult(interp, gspObjCmdEnvfrom);
			break;
		case OPT_ENVRCPT:
			SetObjResult(interp, gspObjCmdEnvrcpt);
			break;
		case OPT_HEADER:
			SetObjResult(interp, gspObjCmdHeader);
			break;
		case OPT_EOH:
			SetObjResult(interp, gspObjCmdEoh);
			break;
		case OPT_BODY:
			SetObjResult(interp, gspObjCmdBody);
			break;
		case OPT_EOM:
			SetObjResult(interp, gspObjCmdEom);
			break;
		case OPT_ABORT:
			SetObjResult(interp, gspObjCmdAbort);
			break;
		case OPT_CLOSE:
			SetObjResult(interp, gspObjCmdClose);
			break;

		default:
			Tcl_SetResult(interp, "Internal error: command not supported yet",
				TCL_STATIC);
			return TCL_ERROR;
		}

		Tcl_MutexUnlock(&gsMutex);
	}
	else
	{
		Tcl_Obj**	ppCfgObj = NULL;

		switch ((enum cmdOptions) optIndex)
		{
		case OPT_CONNECT:
			ppCfgObj = &gspObjCmdConnect;
			break;
		case OPT_HELO:
			ppCfgObj = &gspObjCmdHelo;
			break;
		case OPT_ENVFROM:
			ppCfgObj = &gspObjCmdEnvfrom;
			break;
		case OPT_ENVRCPT:
			ppCfgObj = &gspObjCmdEnvrcpt;
			break;
		case OPT_HEADER:
			ppCfgObj = &gspObjCmdHeader;
			break;
		case OPT_EOH:
			ppCfgObj = &gspObjCmdEoh;
			break;
		case OPT_BODY:
			ppCfgObj = &gspObjCmdBody;
			break;
		case OPT_EOM:
			ppCfgObj = &gspObjCmdEom;
			break;
		case OPT_ABORT:
			ppCfgObj = &gspObjCmdAbort;
			break;
		case OPT_CLOSE:
			ppCfgObj = &gspObjCmdClose;
			break;

		default:
			Tcl_SetResult(interp, "Internal error: command not supported yet",
				TCL_STATIC);
			return TCL_ERROR;
		}

		Tcl_MutexLock(&gsMutex);

		if (ppCfgObj)
		{
			if (*ppCfgObj)
			{
				Tcl_DecrRefCount(*ppCfgObj);
				*ppCfgObj = NULL;
			}

			if (Tcl_GetCharLength(objv[0]))
			{
				*ppCfgObj = objv[0];
				Tcl_IncrRefCount(objv[0]);
			}
		}

		Tcl_MutexUnlock(&gsMutex);
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_StopObjCmd --
 *
 *	This procedure is a handler for the "milter stop" Tcl command, which
 *	is a wrapper around smfi_stop() to shut down the milter.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	smfi_main returns.  As of this writing, smfi_main cannot be invoked
 *	again once it returns (despite documentation to the contrary).  Once
 *	stopped, the Tcl process should exit.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_StopObjCmd
(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	if (objc > 1)
	{
		Tcl_WrongNumArgs(interp, 1, objv, NULL);
		return TCL_ERROR;
	}

	smfi_stop();
	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_StopObjCmd --
 *
 *	This procedure is a handler for the "milter main" Tcl command, which
 *	is a wrapper around smfi_main() to connect to the MTA and start
 *	processing messages.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	This function will never return unless smfi_stop is called by another
 *	thread (e.g. via "milter stop") or the process has received a kill
 *	signal.  As of this writing, smfi_main cannot be invoked
 *	again once it returns (despite documentation to the contrary).  Once
 *	stopped, the Tcl process should exit.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_MainObjCmd
(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	if (objc > 1)
	{
		Tcl_WrongNumArgs(interp, 1, objv, NULL);
		return TCL_ERROR;
	}

	smfilter.xxfi_name = Tcl_GetString(gspObjMilterName);
	smfi_register(smfilter);

	if (smfi_main() == MI_FAILURE)
	{
		Tcl_SetResult(interp, "smfi_main: connection failed", TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_MilterObjCmd --
 *
 *	This procedure is a handler for the "milter" Tcl command.  It's
 *	used to invoke subcommands for configuring and controlling the
 *	milter.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	Depends on subcommand.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_MilterObjCmd
(ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	static const char* ppCmdOptions[] =
	{
		"configure", "main", "register", "stop",
		(char *)NULL
	};

	enum cmdOptions
	{
		CMD_CONFIGURE, CMD_MAIN, CMD_REGISTER, CMD_STOP
	};

	int optIndex;

	if (objc < 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "command ?arg arg ...?");
		return TCL_ERROR;
	}

	objc--; objv++;

	if (Tcl_GetIndexFromObj(interp, objv[0], ppCmdOptions, "command",
			0, &optIndex) == TCL_ERROR)
		return TCL_ERROR;

	if ((((enum cmdOptions) optIndex) != CMD_STOP) && (interp != gpInterp))
	{
		Tcl_SetResult(interp,
			"Only the stop commmand can be called from secondary threads",
			TCL_STATIC);
		return TCL_ERROR;
	}

	switch ((enum cmdOptions) optIndex)
	{
	case CMD_CONFIGURE:
		return TM_ConfigObjCmd(interp, objc, objv);

	case CMD_MAIN:
		return TM_MainObjCmd(interp, objc, objv);

	case CMD_REGISTER:
		return TM_RegisterObjCmd(interp, objc, objv);

	case CMD_STOP:
		return TM_StopObjCmd(interp, objc, objv);
	}

	Tcl_SetResult(interp, "Internal error: command not supported yet",
		TCL_STATIC);

	return TCL_ERROR;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_GetSymVal --
 *
 *	This procedure is a wrapper for the smfi_getsymval function, which
 *	return the value of the specified macro from sendmail.  The macros
 *	that are available to be queried depend on the state of the message
 *	processing (i.e., which callback is being invoked).
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	None
 *
 *-------------------------------------------------------------------------
 */

static int
TM_GetSymVal(SMFICTX* pCtx, Tcl_Interp* interp,
	int objc, Tcl_Obj* CONST objv[])
{
	Tcl_Obj*	pObjResult	= NULL;
	char*		pValue		= NULL;

	if (objc != 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "symbol");
		return TCL_ERROR;
	}

	objc--; objv++;

	pValue = smfi_getsymval(pCtx, Tcl_GetString(objv[0]));

	if (pValue)
		pObjResult = Tcl_NewStringObj(pValue, -1);
	else
		pObjResult = Tcl_NewObj();

	Tcl_SetObjResult(interp, pObjResult);
	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_GetPriv --
 *
 *	This procedure provides read access to a context's private data
 *	hash table.  It behaves essentially the same as "set" without a
 *	value.  If no name is given, it will return a list of name/value
 *	pairs for all of the private data.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	None
 *
 *-------------------------------------------------------------------------
 */

static int
TM_GetPriv(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	TM_PrivData*	pPriv		= NULL;

	if (objc > 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "?name?");
		return TCL_ERROR;
	}

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv)
	{
		Tcl_HashEntry*	pEnt		= NULL;
		Tcl_Obj*		pObjName	= NULL;
		Tcl_Obj*		pObjValue	= NULL;

		if (objc == 1)
		{
			Tcl_Obj*		pObjResult	= Tcl_NewObj();
			Tcl_HashSearch	status;

			pEnt = Tcl_FirstHashEntry(&pPriv->mData, &status);

			while (pEnt)
			{
				pObjName	= (Tcl_Obj*)Tcl_GetHashKey(&pPriv->mData, pEnt),
				pObjValue	= (Tcl_Obj*)Tcl_GetHashValue(pEnt);

				Tcl_ListObjAppendElement(interp, pObjResult, pObjName);
				Tcl_ListObjAppendElement(interp, pObjResult, pObjValue);

				pEnt = Tcl_NextHashEntry(&status);
			}

			Tcl_SetObjResult(interp, pObjResult);
		}
		else
		{
			pEnt = Tcl_FindHashEntry(&pPriv->mData, (char *)objv[1]);

			if (!pEnt)
			{
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "can't read \"",
					Tcl_GetString(objv[1]), "\": no such key", NULL);
				Tcl_MutexUnlock(&gsMutex);
				return TCL_ERROR;
			}

			Tcl_SetObjResult(interp, (Tcl_Obj*)Tcl_GetHashValue(pEnt));
		}
	}

	Tcl_MutexUnlock(&gsMutex);

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_SetPriv --
 *
 *	This procedure provides read-write access to a context's private data
 *	hash table.  It behaves essentially the same as "set" with or without
 *	a value.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	None
 *
 *-------------------------------------------------------------------------
 */

static int
TM_SetPriv(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	TM_PrivData*	pPriv		= NULL;

	if ((objc < 2) || (objc > 3))
	{
		Tcl_WrongNumArgs(interp, 1, objv, "name ?value?");
		return TCL_ERROR;
	}

	if (objc == 2)
		return TM_GetPriv(pCtx, interp, objc, objv);

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv)
	{
		Tcl_HashEntry*	pEnt	= NULL;
		int				newEnt;

		pEnt = Tcl_CreateHashEntry(&pPriv->mData, (char *)objv[1], &newEnt);

		if (pEnt)
		{
			Tcl_IncrRefCount(objv[2]);

			if (!newEnt)
			{
				Tcl_Obj* pObjValue	= (Tcl_Obj*)Tcl_GetHashValue(pEnt);
				Tcl_DecrRefCount(pObjValue);
			}

			Tcl_SetHashValue(pEnt, objv[2]);
		}

		Tcl_SetObjResult(interp, objv[2]);
	}

	Tcl_MutexUnlock(&gsMutex);

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_UnsetPriv --
 *
 *	This procedure provides the ability to unset fields in a context's
 *	private data hash table.  It behaves essentially the same as
 *	"array unset" with or without a pattern.  If no pattern is given,
 *	all private data is unset.  Otherwise, all entries whose names
 *	match the pattern are unset.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	None
 *
 *-------------------------------------------------------------------------
 */

static int
TM_UnsetPriv(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	TM_PrivData*	pPriv		= NULL;

	if (objc > 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "?pattern?");
		return TCL_ERROR;
	}

	Tcl_MutexLock(&gsMutex);

	pPriv = (TM_PrivData *)smfi_getpriv(pCtx);

	if (pPriv)
	{
		Tcl_HashEntry*	pEnt		= NULL;
		Tcl_HashEntry*	pNextEnt	= NULL;
		char*			pPattern	= NULL;
		Tcl_Obj*		pObjKey		= NULL;
		Tcl_HashSearch	status;

		if (objc == 2)
			pPattern = Tcl_GetString(objv[1]);

		pEnt = Tcl_FirstHashEntry(&pPriv->mData, &status);

		while (pEnt)
		{
			pNextEnt	= Tcl_NextHashEntry(&status);
			pObjKey		= (Tcl_Obj*)Tcl_GetHashKey(&pPriv->mData, pEnt);

			if (!pPattern || Tcl_StringMatch(Tcl_GetString(pObjKey), pPattern))
			{
				Tcl_DecrRefCount((Tcl_Obj*)Tcl_GetHashValue(pEnt));
				Tcl_DeleteHashEntry(pEnt);
			}

			pEnt = pNextEnt;
		}
	}

	Tcl_MutexUnlock(&gsMutex);

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_SetReply --
 *
 *	This procedure provides a wrapper to the smfi_setreply() function,
 *	allowing extended SMTP result codes and messages to be provided for
 *	temporary and permanent failures.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The SMTP result code and details for the message may be changed.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_SetReply(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pRCode	= NULL;
	char*	pXCode	= NULL;
	char*	pMsg	= NULL;

	if ((objc < 2) || (objc > 4))
	{
		Tcl_WrongNumArgs(interp, 1, objv, "rcode ??xcode? message?");
		return TCL_ERROR;
	}

	objc--; objv++;

	pRCode	= Tcl_GetString(objv[0]);
	objc--; objv++;

	if (objc > 1)
	{
		pXCode	= Tcl_GetString(objv[0]);
		objc--; objv++;
	}

	if (objc)
		pMsg	= Tcl_GetString(objv[0]);

	if (smfi_setreply(pCtx, pRCode, pXCode, pMsg) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_setreply failed: bad rcode/xcode or memory allocation failed",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_AddHeader --
 *
 *	This procedure provides a wrapper to the smfi_addheader() function,
 *	allowing headers to be added to the message.  Each header is added
 *	to the end, after all existing headers.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message is rewritten to add headers.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_AddHeader(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pName	= NULL;
	char*	pValue	= NULL;

	if (objc != 3)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "name data");
		return TCL_ERROR;
	}

	pName	= Tcl_GetString(objv[1]);
	pValue	= Tcl_GetString(objv[2]);

	if (smfi_addheader(pCtx, pName, pValue) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_addheader failed: called outside eom or memory allocation failed",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_ChgHeader --
 *
 *	This procedure provides a wrapper to the smfi_chgheader() function,
 *	allowing existing headers to rewritten.  If the header doesn't exist,
 *	it is added to the end, after all existing headers.  An index can
 *	be specified to indicate which of several headers of the same type
 *	is to be changed, starting at 1.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message is rewritten to change or add headers.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_ChgHeader(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pName	= NULL;
	char*	pValue	= NULL;
	int		index	= 1;

	if ((objc < 3) || (objc > 4))
	{
		Tcl_WrongNumArgs(interp, 1, objv, "name ?index? data");
		return TCL_ERROR;
	}

	objc--; objv++;

	pName	= Tcl_GetString(objv[0]);
	objc--; objv++;

	if (objc > 1)
	{
		if (Tcl_GetIntFromObj(interp, objv[0], &index) == TCL_ERROR)
			return TCL_ERROR;

		objc--; objv++;
	}

	pValue	= Tcl_GetString(objv[0]);

	if (smfi_chgheader(pCtx, pName, index, pValue) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_chgheader failed: called outside eom or memory allocation failed",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_InsHeader --
 *
 *	This procedure provides a wrapper to the smfi_insheader() function,
 *	allowing headers to be inserted before or after existing headers.
 *	The header is inserted at the specified index, if any, or as the
 *	first header otherwise.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message is rewritten to add headers.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_InsHeader(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pName	= NULL;
	char*	pValue	= NULL;
	int		index	= 0;

	if ((objc < 3) || (objc > 4))
	{
		Tcl_WrongNumArgs(interp, 1, objv, "name ?index? data");
		return TCL_ERROR;
	}

	objc--; objv++;

	pName	= Tcl_GetString(objv[0]);
	objc--; objv++;

	if (objc > 1)
	{
		if (Tcl_GetIntFromObj(interp, objv[0], &index) == TCL_ERROR)
			return TCL_ERROR;

		objc--; objv++;
	}

	pValue	= Tcl_GetString(objv[0]);

	if (smfi_insheader(pCtx, index, pName, pValue) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_insheader failed: called outside eom or memory allocation failed",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_ReplaceBody --
 *
 *	This procedure provides a wrapper to the smfi_replacebody() function,
 *	allowing the body of a message to be completely rewritten.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message body is rewritten.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_ReplaceBody(SMFICTX* pCtx, Tcl_Interp* interp,
	int objc, Tcl_Obj* CONST objv[])
{
	char*	pBody	= NULL;
	int		bodyLen	= 0;

	if (objc != 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "data");
		return TCL_ERROR;
	}

	objc--; objv++;

	pBody	= Tcl_GetByteArrayFromObj(objv[0], &bodyLen);

	if (smfi_replacebody(pCtx, pBody, bodyLen) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_replacebody failed: cannot change body in current state or network error occurred.",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_AddRcpt --
 *
 *	This procedure provides a wrapper to the smfi_addrcpt() function,
 *	allowing the message to be copied to additional recipients.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message is sent to additional recipients.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_AddRcpt(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pRcpt	= NULL;

	if (objc != 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "rcpt");
		return TCL_ERROR;
	}

	objc--; objv++;

	pRcpt	= Tcl_GetString(objv[0]);

	/* TODO: what happens if pRcpt is an empty string? */

	if (smfi_addrcpt(pCtx, pRcpt) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_addrcpt failed: cannot change add recipients in current state or network error occurred.",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_DelRcpt --
 *
 *	This procedure provides a wrapper to the smfi_delrcpt() function,
 *	allowing recipients to be removed from the set that the message will
 *	be sent to.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The message is NOT sent to deleted recipients.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_DelRcpt(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	char*	pRcpt	= NULL;

	if (objc != 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "rcpt");
		return TCL_ERROR;
	}

	objc--; objv++;

	pRcpt	= Tcl_GetString(objv[0]);

	/* TODO: what happens if pRcpt is an empty string? */

	if (smfi_delrcpt(pCtx, pRcpt) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_delrcpt failed: cannot change add recipients in current state or network error occurred.",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_Progress --
 *
 *	This procedure provides a wrapper to the smfi_progress() function to
 *	reset timeouts and indicate that the filter is still processing the
 *	message.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	Timer is reset.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_Progress(SMFICTX* pCtx, Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
{
	if (objc != 1)
	{
		Tcl_WrongNumArgs(interp, 1, objv, NULL);
		return TCL_ERROR;
	}

	if (smfi_progress(pCtx) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_progress failed: network error occurred.",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_Quarantine --
 *
 *	This procedure provides a wrapper to the smfi_quarantine() function to
 *	indicate that the message should be saved in the queue but not
 *	delivered without administrative intervention.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	Timeouts are reset.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_Quarantine(SMFICTX* pCtx, Tcl_Interp* interp,
	int objc, Tcl_Obj* CONST objv[])
{
	if (objc != 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "reason");
		return TCL_ERROR;
	}

	if (smfi_quarantine(pCtx, Tcl_GetString(objv[1])) == MI_FAILURE)
	{
		Tcl_SetResult(interp,
			"smfi_quarantine failed: network error occurred.",
			TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_CtxObjCmd --
 *
 *	This procedure provides the handler for each context's associated
 *	Tcl command, allowing messages to be rewritten, rejected, quarantined,
 *	sent to more or fewer people, etc.  The command is created at connection
 *	time and deleted during message cleanup.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	Depends on the subcommand.
 *
 *-------------------------------------------------------------------------
 */

static int
TM_CtxObjCmd(ClientData clientData, Tcl_Interp* interp,
	int objc, Tcl_Obj* CONST objv[])
{
	SMFICTX* pCtx	= (SMFICTX *)clientData;

	static const char* ppCmdOptions[] =
	{
		"getsymval", "getpriv", "setpriv", "unsetpriv", "setreply",
		"addheader", "chgheader", "insheader", "addrcpt",
		"delrcpt", "replacebody", "progress", "quarantine",
		(char *)NULL
	};

	enum cmdOptions
	{
		CMD_GETSYMVAL, CMD_GETPRIV, CMD_SETPRIV, CMD_UNSETPRIV, CMD_SETREPLY,
		CMD_ADDHEADER, CMD_CHGHEADER, CMD_INSHEADER, CMD_ADDRCPT,
		CMD_DELRCPT, CMD_REPLACE, CMD_PROGRESS, CMD_QUARANTINE
	};

	int optIndex;

	if (objc < 2)
	{
		Tcl_WrongNumArgs(interp, 1, objv, "command ?arg arg ...?");
		return TCL_ERROR;
	}

	objc--; objv++;

	if (Tcl_GetIndexFromObj(interp, objv[0], ppCmdOptions, "command",
			0, &optIndex) == TCL_ERROR)
		return TCL_ERROR;

	switch ((enum cmdOptions) optIndex)
	{
	case CMD_GETSYMVAL:
		return TM_GetSymVal(pCtx, interp, objc, objv);
	case CMD_GETPRIV:
		return TM_GetPriv(pCtx, interp, objc, objv);
	case CMD_SETPRIV:
		return TM_SetPriv(pCtx, interp, objc, objv);
	case CMD_UNSETPRIV:
		return TM_UnsetPriv(pCtx, interp, objc, objv);
	case CMD_SETREPLY:
		return TM_SetReply(pCtx, interp, objc, objv);
	case CMD_ADDHEADER:
		return TM_AddHeader(pCtx, interp, objc, objv);
	case CMD_CHGHEADER:
		return TM_ChgHeader(pCtx, interp, objc, objv);
	case CMD_INSHEADER:
		return TM_InsHeader(pCtx, interp, objc, objv);
	case CMD_ADDRCPT:
		return TM_AddRcpt(pCtx, interp, objc, objv);
	case CMD_DELRCPT:
		return TM_DelRcpt(pCtx, interp, objc, objv);
	case CMD_REPLACE:
		return TM_ReplaceBody(pCtx, interp, objc, objv);
	case CMD_PROGRESS:
		return TM_Progress(pCtx, interp, objc, objv);
	case CMD_QUARANTINE:
		return TM_Quarantine(pCtx, interp, objc, objv);
	}
	
	Tcl_SetResult(interp, "Internal error: command not supported yet",
		TCL_STATIC);

	return TCL_ERROR;
}


/*
 *-------------------------------------------------------------------------
 *
 * TM_MilterObjCmdDeleted --
 *
 *	This procedure is invoked by Tcl when the "milter" command is deleted.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The milter is stopped and "milter main" returns.
 *
 *-------------------------------------------------------------------------
 */

static void
TM_MilterObjCmdDeleted(ClientData clientData)
{
	smfi_stop();
}


/*
 *-------------------------------------------------------------------------
 *
 * Tclmilter_SafeInit --
 *
 *	This procedure is invoked by Tcl when the package is loaded in a safe
 *	interpreter.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The "milter" command is provided to the interpreter.
 *
 *-------------------------------------------------------------------------
 */

int
Tclmilter_SafeInit(Tcl_Interp* interp)
{
	/*
	 * Check minimum Tcl requirements and initialize stubs table.
	 */

	if (Tcl_InitStubs(interp, "8.1", 0) == NULL)
		return TCL_ERROR;

	if (Tcl_PkgRequire(interp, "Tcl", "8.1", 0) == NULL)
		return TCL_ERROR;

	/*
	 * If this is the first time the package is being loaded,
	 * stash away the interpreter to be used as the "primary" interpreter,
	 * and initialize the milter name and connection to default.
	 */

	if (!gpInterp)
	{
		gspObjMilterName = Tcl_NewStringObj("TclMilter", 9);;
		gspObjConnection = Tcl_NewStringObj("unix:/var/run/tclmilter.sock", 23);

		Tcl_IncrRefCount(gspObjMilterName);
		Tcl_IncrRefCount(gspObjConnection);

		gpInterp = interp;
	}

	/*
	 * Make the package and milter command available.
	 */

	if (Tcl_CreateObjCommand(interp, "milter", TM_MilterObjCmd,
			(ClientData)NULL, TM_MilterObjCmdDeleted) == NULL)
		return TCL_ERROR;

	if (Tcl_PkgProvide(interp, "TclMilter", PACKAGE_VERSION) != TCL_OK)
		return TCL_ERROR;

	return TCL_OK;
}


/*
 *-------------------------------------------------------------------------
 *
 * Tclmilter_Init --
 *
 *	This procedure is invoked by Tcl when the package is loaded in a
 *	normal interpreter.  Since this package's its self doesn't do anything
 *	unsafe beyond connection to the MTA through libmilter, this just
 *	invokes the safe init function.
 *
 * Results:
 * 	TCL_OK
 * 	TCL_ERROR
 *
 * Side effects:
 *	The "milter" command is provided to the interpreter.
 *
 *-------------------------------------------------------------------------
 */

int
Tclmilter_Init(Tcl_Interp* interp)
{
	return Tclmilter_SafeInit(interp);
}

